# coding=utf-8
#
# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis-python
#
# Most of this work is copyright (C) 2013-2018 David R. MacIver
# (david@drmaciver.com), but it contains contributions by others. See
# CONTRIBUTING.rst for a full list of people who may hold copyright, and
# consult the git log if you need to determine who owns an individual
# contribution.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at http://mozilla.org/MPL/2.0/.
#
# END HEADER

from __future__ import division, print_function, absolute_import

from tests.common.utils import capture_out
from hypothesis.internal.conjecture.shrinking import Integer


def test_debug_output():
    with capture_out() as o:
        Integer.shrink(10, lambda x: True, debug=True)

    assert 'initial=10' in o.getvalue()
    assert 'shrinking to 0' in o.getvalue()


def test_includes_name_in_repr_if_set():
    assert repr(
        Integer(10, lambda x: True, name='hi there')
    ) == "Integer('hi there', initial=10, current=10)"


def test_normally_contains_no_space_for_name():
    assert repr(
        Integer(10, lambda x: True)) == 'Integer(initial=10, current=10)'
